import numpy as np
import numpy.matlib as npm
import matplotlib.pyplot as plt


def gps_generator(num_points, r=1000.0, s=0.0, k=100.0, w=40.0):
    """
    Generates artificial GPS measurements of a vehicle that drives over a modulated circular trajectory.

    :param int num_points: Number of trajectory points to generate
    :param float r: Base radius of the trajectory circle
    :param float s: Standard deviation of the normally distributed GPS scatter
    :param k: Amplitude of radius modulation (so that radius goes from `r-k` to `r+k`)
    :param w: Angular frequency of radius modulation
    :return array: NumPy array `(2,num_points)`
    """

    # Finish after one round
    theta_stop = 2*np.pi
    # Pre-compute the angular step so that we can use np.arange()
    theta_step = theta_stop/num_points

    # Pre-allocate space for output
    coords = np.zeros([2, num_points])
    # Generate the GPS scatter in one sweep so that we do not have to generate it element-by-element
    # in the main loop
    randns = np.random.standard_normal(coords.shape)

    # Initialise the column pointer
    col = 0
    # Note: np.arange() has a different order of parameters than the standard range().
    # Note: One may try np.linspace() as well.
    for theta in np.arange(0, theta_stop, theta_step):
        # Modulated radius
        q = (r + k * np.sin(w * theta))
        # True modulated coordinates
        x = q * np.cos(theta)
        y = q * np.sin(theta)
        # And possibly some additional noise to account for GPS signal scatter
        x = x + randns[0, col] * s
        y = y + randns[1, col] * s
        # NumPy allows us to store list into an array axis
        coords[:, col] = [x, y]
        # Move the column pointer
        col += 1

    return coords


def kalman(x0, u0, ym, A, B, C, D, P, Q, R):
    """
    Simple Kalman filter for a LTI system.

    :param matrix x0: Current state, column vector represented as (n,1) matrix
    :param matrix u0: Current input, column vector represented as (i,1) matrix
    :param matrix ym: Current measurements, column vector represented as (m,1) matrix
    :param matrix A: State development matrix, (n,n)
    :param matrix B: Input scaling matrix, (n,i)
    :param matrix C: Output matrix, (m,n), may be time-variable
    :param matrix D: Intercept matrix, (m,i), typically empty
    :param matrix P: A-posteriori estimate of state covariance x0
    :param matrix Q: Process noise covariance
    :param matrix R: Observation noise covariance
    :return tuple: (x0f,x1p,Pf,P) filtered and predicted state and covariances
    """

    # -----------
    # Update step
    # -----------
    # Output prediction
    yp = C*x0 + D*u0
    # Innovation or measurement residual
    ey = ym - yp
    # Innovation covariance
    S = C*P*C.T + R
    # Optimal Kalman gain
    # We try to avoid computing the inverse of S -- the number of transposes
    # is due to conversion from Matlab to NumPy ...
    K = np.linalg.solve(S.T, C*P.T).T
    # Original: K = P*C'*inv(S);
    # Updated (a posteriori) state estimate
    x0f = x0 + K*ey
    # Updated (a posteriori) estimate covariance
    Pf = P - K*C*P

    # ---------------
    # Prediction step
    # ---------------
    # New predicted state
    x1p = A*x0f + B*u0
    # The predicted (a priori) estimate covariance is the filtered covariance
    # scaled by the state development matrix + process noise
    P = A*Pf*A.T + Q

    return (x0f, x1p, Pf, P)


# How many syhtetic GPS points shall be generated
num_points = 300

# Generate GPS points
data = gps_generator(num_points, s=10, w=10)
# Data is an ndarray which means that all typical operations will be carried out element-wise.
# We will convert it to matrix
mdata = np.asmatrix(data)

# **********************************************************************************
#  Simple vehicle model that assumes static vehicle
# **********************************************************************************
if False:
    # State development: x[n+1] = x[n]
    # Output: y[n] = x[n]
    # This model is practically useless!
    A = np.matrix([[1, 0], [0, 1]], dtype=float)
    B = np.matrix([[0], [0]], dtype=float)
    C = np.matrix([[1, 0], [0, 1]], dtype=float)
    D = np.matrix([[0], [0]], dtype=float)

    # -------------------
    # Initial covariances
    # -------------------
    # Initial state noise, decreases exponentially
    P0 = np.asmatrix(np.diag([10, 10]))

    # Different filtering strategies are parametrised by process and measurement noise covariances
    # (1) We believe in the model
    # Q = np.asmatrix(np.diag([0.00001, 0.00001]))    % process noise
    # R = np.asmatrix(np.diag([100, 100]))            % measurement noise
    # (2) We believe in measurement
    Q = np.asmatrix(np.diag([1, 1]))  # process noise
    R = np.asmatrix(np.diag([1, 1]))  # measurement noise

    # Initial vehicle position
    x0 = mdata[:, 0]
    # Allocate space for recording the kalman-filtered model predictions
    filt_history = np.asmatrix(np.zeros([2, num_points]))
    filt_history[:, 0] = x0
    # Allocate space for recording unfiltered model predictions
    pred_history = np.asmatrix(np.zeros([2, num_points]))
    pred_history[:, 0] = x0

    # And now step over all simulated GPS measurements
    for i in range(1, num_points):
        # Measurement
        ym = mdata[:, i]
        # Filter the predicted state at this moment using Kalman filter
        (x0f, x1p, P0f, P1p) = kalman(x0, 0, ym, A, B, C, D, P0, Q, R)
        # Record the filtered state
        filt_history[:, i] = x0f
        # Record the model prediction, i.e. the position of the vehicle in the next step
        pred_history[:, i] = x1p
        # The updated state will be the "old" state in the next step
        x0 = x1p
        # The updated state noise will be the "old" state noise in the next step
        P0 = P1p

    plt.figure(1)
    # Note: matplotlib.pyplot.plot() plots matrix _columns_, hence `filt_history` has to be transposed
    plt.plot(data[0, :], data[1, :], '.', filt_history[0, :].T, filt_history[1, :].T, '-')
    plt.legend(['GPS position', 'Kalman-filtered'])
    plt.xlabel('Vehicle x coordinate [m]')
    plt.ylabel('Vehicle y coordinate [m]')
    plt.show()

# **********************************************************************************
# Slightly more ambitious model, assuming constant movement
# **********************************************************************************
if True:
    # State development (xy0 is the previous position, xy1 is the current one):
    #     xy0[n+1] = xy1[n]
    #     xy1[n+1] = xy1[n]+(xy1[n]-xy0[n]) = 2*xy1[n]-xy0[n]
    # Output:
    #     y[n] = xy1[n]
    A = np.matrix([[0, 0, 1, 0], [0, 0, 0, 1], [-1, 0, 2, 0], [0, -1, 0, 2]], dtype=float)
    B = np.matrix([[0], [0], [0], [0]], dtype=float)
    C = np.matrix([[0, 0, 1, 0], [0, 0, 0, 1]], dtype=float)
    D = np.matrix([[0], [0]], dtype=float)

    # -------------------
    # Initial covariances
    # -------------------
    # Initial state noise, decreases exponentially
    P0 = np.asmatrix(np.diag([10, 10, 10, 10]))

    # Different filtering strategies are parametrised by process and measurement noise covariances
    # (1) We believe in the model
    # Q = np.asmatrix(np.diag([0.00001, 0.00001]))    % process noise
    # R = np.asmatrix(np.diag([100, 100]))            % measurement noise
    # (2) We believe in measurement
    Q = np.asmatrix(np.diag([1, 1, 1, 1]))          # process noise
    R = np.asmatrix(np.diag([1, 1]))      # measurement noise

    # Initial vehicle position
    x0 = np.concatenate((mdata[:, 0], mdata[:, 0]))
    # Allocate space for recording the kalman-filtered model predictions
    filt_history = np.asmatrix(np.zeros([4, num_points]))
    filt_history[:, 0] = x0
    # Allocate space for recording unfiltered model predictions
    pred_history = np.asmatrix(np.zeros([4, num_points]))
    pred_history[:, 0] = x0

    # And now step over all simulated GPS measurements
    for i in range(1, num_points):
        # Measurement
        ym = mdata[:, i]
        # Filter the predicted state at this moment using Kalman filter
        (x0f, x1p, P0f, P1p) = kalman(x0, 0, ym, A, B, C, D, P0, Q, R)
        # Record the filtered state
        filt_history[:, i] = x0f
        # Record the model prediction, i.e. the position of the vehicle in the next step
        pred_history[:, i] = x1p
        # The updated state will be the "old" state in the next step
        x0 = x1p
        # The updated state noise will be the "old" state noise in the next step
        P0 = P1p

    plt.figure(1)
    # Note: matplotlib.pyplot.plot() plots matrix _columns_, hence `filt_history` has to be transposed
    plt.plot(data[0, :], data[1, :], '.', pred_history[2, :].T, pred_history[3, :].T, '-', filt_history[2, :].T, filt_history[3, :].T, '-')
    plt.legend(['GPS position', 'Predicted', 'Kalman-filtered'])
    plt.xlabel('Vehicle x coordinate [m]')
    plt.ylabel('Vehicle y coordinate [m]')
    plt.savefig('demo2.pdf')
    plt.show()
